#!/usr/bin/perl

######################################################################################################
# MP3 Genre Sorter
######################################################################################################
# 2003 by Henning Koch (jaz@netalive.org)
# Use on your own risk
# For private use only
# Please leave this notice intact when redistributing the program.
######################################################################################################


use strict;
use MP3::Info;
use File::Copy;
use POSIX qw(locale_h);

setlocale(LC_CTYPE, 'de_DE.ISO8859-1');
our $DIR_SEPERATOR = "\\";

our %GENRES = (
	'Ballads'			=> 'Ballads',
	'Comedy'			=> 'Comedy',
	'Dance'				=> 'Dance',
	'Electronic'			=> 'Electronic',
	'Classical'			=> 'Classical',
	'World'				=> 'World',
	'Jazz'				=> 'Jazz',
	'Other'				=> 'Other',
	'Black Beat'			=> 'Black Beat',
	'Pop'				=> 'Pop',
	'Rock'				=> 'Rock',
	'Reggae'			=> 'Reggae',
	'Soundtrack'			=> 'Soundtrack',
	'Spoken Word'			=> 'Spoken Word',
);
	
our %GENRE_HOTKEYS = (
	'Ballads'			=> 'B',
	'Comedy'			=> 'M',
	'Dance'				=> 'D',
	'Electronic'			=> 'E',
	'Classical'			=> 'C',
	'World'				=> 'W',
	'Jazz'				=> 'J',
	'Other'				=> 'O',
	'Black Beat'			=> 'K',
	'Pop'				=> 'P',
	'Rock'				=> 'R',
	'Reggae'			=> 'G',
	'Soundtrack'			=> 'S',
	'Spoken Word'			=> 'N',
);	
	


our %GENRE_MAPPINGS = (

	'Acoustic'			=> 'Ballads',
	'Ballad'			=> 'Ballads',
	'Blues'				=> 'Ballads',
	'Easy Listening'		=> 'Ballads',
	'Power Ballad'			=> 'Ballads',

	'Comedy'			=> 'Comedy',
	'Humour'			=> 'Comedy',
	'Opera'				=> 'Comedy',
	'Pranks'			=> 'Comedy',
	'Satire'			=> 'Comedy',

	'Acid'				=> 'Dance',
	'Cabaret'			=> 'Dance',
	'Club'				=> 'Dance',
	'Dance'				=> 'Dance',
	'Dance Hall'			=> 'Dance',
	'Disco'				=> 'Dance',
	'Dream'				=> 'Dance',
	'Euro-House'			=> 'Dance',
	'Euro-Techno'			=> 'Dance',
	'Eurodance'			=> 'Dance',
	'Funk'				=> 'Dance',
	'Hardcore'			=> 'Dance',
	'House'				=> 'Dance',
	'Jungle'			=> 'Dance',
	'Rave'				=> 'Dance',
	'Techno'			=> 'Dance',
	'Techno-Industrial'		=> 'Dance',
	'Trance'			=> 'Dance',

	'Ambient'			=> 'Electronic',
	'Electronic'			=> 'Electronic',
	'Ethnic'			=> 'Electronic',
	'Freestyle'			=> 'Electronic',
	'Industrial'			=> 'Electronic',
	'New Age'			=> 'Electronic',
	'Psychadelic'			=> 'Electronic',
	'Trip-Hop'			=> 'Electronic',

	'A capella'			=> 'Classical',
	'Chorus'			=> 'Classical',
	'Classical'			=> 'Classical',
	'Drum Solo'			=> 'Classical',
	'Duet'				=> 'Classical',
	'Instrumental'			=> 'Classical',
	'Meditative'			=> 'Classical',
	'Musical'			=> 'Classical',
	'Sonata'			=> 'Classical',
	'Symphony'			=> 'Classical',
	'Vocal'				=> 'Classical',

	'Bluegrass'			=> 'World',
	'Celtic'			=> 'World',
	'Chamber Music'			=> 'World',
	'Cult'				=> 'World',
	'Folk'				=> 'World',
	'Folk-Rock'			=> 'World',
	'Folklore'			=> 'World',
	'National Folk'			=> 'World',
	'Native American'		=> 'World',
	'Polka'				=> 'World',
	'Rhythmic Soul'			=> 'World',
	'Samba'				=> 'World',
	'Tango'				=> 'World',
	'Tribal'			=> 'World',

	'Acid Jazz'			=> 'Jazz',
	'Bebob'				=> 'Jazz',
	'Big Band'			=> 'Jazz',
	'Jazz'				=> 'Jazz',
	'Jazz+Funk'			=> 'Jazz',
	'Swing'				=> 'Jazz',

	'Lo-Fi'				=> 'Other',
	'Misc'				=> 'Other',
	'Noise'				=> 'Other',
	'Other'				=> 'Other',
	'Sound Clip'			=> 'Other',
	'Space'				=> 'Other',

	'Christian Rap'			=> 'Black Beat',
	'Gangsta'			=> 'Black Beat',
	'Hip-Hop'			=> 'Black Beat',
	'R&B'				=> 'Black Beat',
	'Rap'				=> 'Black Beat',
	'Soul'				=> 'Black Beat',

	'Country'			=> 'Pop',
	'Gospel'			=> 'Pop',
	'Instrumental Pop'		=> 'Pop',
	'Latin'				=> 'Pop',
	'New Wave'			=> 'Pop',
	'Oldies'			=> 'Pop',
	'Pop'				=> 'Pop',
	'Pop-Folk'			=> 'Pop',
	'Pop/Funk'			=> 'Pop',
	'Porn Groove'			=> 'Pop',
	'RB'				=> 'Pop',
	'Retro'				=> 'Pop',
	'Rock & Roll'			=> 'Pop',
	'Top 40'			=> 'Pop',

	'Reggae'			=> 'Reggae',
	'Ska'				=> 'Reggae',

	'Acid Punk'			=> 'Rock',
	'AlternRock'			=> 'Rock',
	'Alternative'			=> 'Rock',
	'Classic Rock'			=> 'Rock',
	'Death Metal'			=> 'Rock',
	'Gothic'			=> 'Rock',
	'Gothic Rock'			=> 'Rock',
	'Grunge'			=> 'Rock',
	'Hard Rock'			=> 'Rock',
	'Instrumental Rock'		=> 'Rock',
	'Metal'				=> 'Rock',
	'Progressive Rock'		=> 'Rock',
	'Psychedelic Rock'		=> 'Rock',
	'Punk'				=> 'Rock',
	'Punk Rock'			=> 'Rock',
	'Rock'				=> 'Rock',
	'Slow Rock'			=> 'Rock',
	'Southern Rock'			=> 'Rock',
	'Symphonic Rock'		=> 'Rock',

	'Darkwave'			=> 'Soundtrack',
	'Game'				=> 'Soundtrack',
	'Showtunes'			=> 'Soundtrack',
	'Soundtrack'			=> 'Soundtrack',
	'Trailer'			=> 'Soundtrack',

	'Radio drama'			=> 'Spoken Word',
	'Radio play'			=> 'Spoken Word',
	'Speech'			=> 'Spoken Word',
	'Spoken word'			=> 'Spoken Word',
	
);



our @STUPID_GENRES = (
	'Genre',
	'Default'
);



our @GENRES = lc_sort(keys %GENRES);
our %PREVIOUS_CHOICES;

my($script_path, $script_name) = split_path($0);
our $PREVIOUS_CHOICES_DATA_PATH = ($script_path eq ''? '.' : $script_path) . '/genre-sorter_choices.txt';




my($collection, $target_collection) = @ARGV;



print "\n\n";
print "*******************************************************************************\n";
print "* MP3 GENRE SORTER                                                            *\n";
print "* 2003 by henning koch (jaz\@netalive.org)                                    *\n";
print "* use on your own risk                                                        *\n";
print "* for private use only                                                        *\n";
print "* www.netalive.org                                                            *\n";
print "*******************************************************************************\n\n";


print 	"USAGE\n\n" .
	"   perl genre-sorter.pl [COLLECTION] [NEW_COLLECTION]\n\n" .

	"PARAMETERS\n\n" .

	"   COLLECTION       the path to your MP3 collection\n" .
	"                    (optional, default is the current directory)\n\n" .

	"   NEW_COLLECTION   the path to your new MP3 collection\n" .
	"                    (optional, default is the current directory)\n\n" .

	"EXAMPLE\n\n" .
	
	"   perl genre_sorter.pl d:/media/mp3\n\n";




$collection = '.' unless $collection ne '';
$target_collection = $collection unless $target_collection ne '';


foreach my $genre (@GENRES) {
	my $path = $collection . '/' . $genre;
	unless(-d $path) {
		error('Directory ' . $path . ' not found. Please create it.');
	}
	
}

for(keys %GENRE_HOTKEYS) {
	$GENRE_HOTKEYS{$_} = lc $GENRE_HOTKEYS{$_};
}


my @hotkeys = @GENRE_HOTKEYS{@GENRES};
my $hotkeys_regexp = lc join('', @hotkeys);

# print $hotkeys_regexp;
# exit;

my %hotkeys;
my $cc = 0;
foreach my $genre (@GENRES) {
	my $key = lc $GENRE_HOTKEYS{$genre};
	$hotkeys{$key} = $cc;
	$cc++;
}


# print join(" ", %hotkeys);
# exit;

#foreach my $key (keys %hotkeys) {
#	print "KEY $key\n";
#	print "IDX $hotkeys{$key}\n";
#	print "GNR $GENRES[$hotkeys{$key}]\n";
#}

# exit;

print "Recalling previous choices... ";

my $previous_choices_data = read_file($PREVIOUS_CHOICES_DATA_PATH);
my $count_previous_choices = 0;

# print $PREVIOUS_CHOICES_DATA_PATH;
# exit;

foreach my $row (split_table($previous_choices_data)) {
	my($artist, $genre) = split_row($row);
	if($artist ne '' && defined $GENRES{$genre}) {
		$PREVIOUS_CHOICES{lc $artist} = $genre;
		$count_previous_choices++;
	}
}

print "$count_previous_choices found\n\n";


print "Browsing MP3 collection... ";

my @files = lc_sort(directory_listing($collection, 0, 0, '*.mp3', 0));
my $total_files = $#files + 1;

print "$total_files files found\n\n";

my $count = 0;

foreach my $file (@files) {

	$count++;

	$file = $collection . '/' . $file;

	my($path, $name) = split_path($file);
	
	$name =~ /^(.+?) - /;
	my $artist = $1;

	my $tag;
	my $genre;

 	$tag = get_mp3tag($file, 1);
 	my $id2_genre_bak = preprocess_genre($tag->{GENRE});
 	my $id2_genre = process_genre($id2_genre_bak);

 	$tag = get_mp3tag($file, 2);
 	my $id1_genre_bak  = preprocess_genre($tag->{GENRE});
 	my $id1_genre = process_genre($id1_genre_bak);
	
	$genre = defined $id2_genre? $id2_genre : $id1_genre;
	
	unless(defined $genre) {
	 	$id2_genre = process_genre_fuzzy($id2_genre_bak);
	 	$id1_genre = process_genre_fuzzy($id1_genre_bak);
	 	$genre = defined $id2_genre? $id2_genre : $id1_genre;
	}
	
	if($artist ne '' && defined $PREVIOUS_CHOICES{lc $artist}) {
		$genre = $PREVIOUS_CHOICES{lc $artist};
	}
	
	my $display = $name . '  [' . $count . '/' . $total_files . ']';

	print "+-----------------------------------------------------------------------------+\n";
	print "| " . pad_right($display, 75) . " |\n";
	print "+-----------------------------------------------------------------------------+\n";
	
	if(!defined $genre && ($id1_genre_bak ne '' || $id2_genre_bak ne '')) {
		print "  $id1_genre_bak?\n" if $id1_genre_bak ne '';
		print "  $id2_genre_bak?\n" if $id2_genre_bak ne '';
		print "+-----------------------------------------------------------------------------+\n";
	}
	
	print "\n";

	my $suggested = undef;
	my $cc = 0;

	foreach my $sort_genre (@GENRES) {
		if($sort_genre eq $genre) {
			$suggested = lc $hotkeys[$cc];
			print "==> ";
		}
		else {
			print "    ";
		}

		print "[" . $hotkeys[$cc] . "] " . $sort_genre . "\n";
		$cc++;
	}

	print "    [#] Skip file\n";
	print "    [.] Quit\n";
	print "    [ENTER] Use suggested genre\n" if defined $genre;

	print "\n";
	print "Select a genre: ";
	
	my $key = '__'; 
	my $allow_enter = (defined $genre? '|' : '');
	
	
	while($key !~ /^ ( [$hotkeys_regexp  \. \# ] $allow_enter ) $/ix) {
		$key = read_line_from_keyboard();
		#print "KEY $key!!\n";
		#print "I WANT: [$hotkeys_regexp] $allow_enter\n";
	}
	
	print "\n";
	
	if($key eq '') {
		$key = $suggested;
	}
	
	if($key eq '#') {
		next;		
	}
	elsif($key eq '.') {
		quit();
	}
	else {
		my $choice = $GENRES[$hotkeys{$key}];
		
		if($artist ne '') {
			$PREVIOUS_CHOICES{lc $artist} = $choice;
		}
		
		my $target_dir = $target_collection . $DIR_SEPERATOR . $choice;
		my $target_file = $target_collection . $DIR_SEPERATOR . $choice . $DIR_SEPERATOR . $name;

		print "Moving file to $target_dir ... ";
		
		move($file, $target_file);
		unless(-f $target_file) {
			error("Couldn't move $file");
		}
		
		print "done\n\n";
	}
	
	
	

}


quit();



sub preprocess_genre {

	my($genre) = @_;
	
	$genre =~ s/[^a-zA-Z0-9\_\-\ ]//g;
	
	$genre = shrink($genre);
	foreach my $stupid (@STUPID_GENRES) {
		if(lc $genre eq lc $stupid) {
			return undef;
		}
	}
	
	return $genre;

}

sub process_genre {

	my($genre) = @_;
	
	# lc vergleichen!
	
	foreach my $test (@GENRES) {
		if ($genre =~ /^$test$/i) { return $test; }
	}

	foreach my $test (keys %GENRE_MAPPINGS) {
		if ($genre =~ /^$test$/i) { return $test; }
	}
	
	return undef;

}


sub process_genre_fuzzy {

	my($genre) = @_;
	
	# lc vergleichen!
	
	foreach my $test (@GENRES) {
		if ($genre =~ /$test/i) { return $test; }
	}

	foreach my $test (keys %GENRE_MAPPINGS) {
		if ($genre =~ /$test/i) { return $test; }
	}
	
	return undef;

}

sub quit {

	my $new_data = '';

	print "Saving choices for future sessions... ";
	foreach my $artist (lc_sort(keys %PREVIOUS_CHOICES)) {
		$new_data .= make_row($artist, $PREVIOUS_CHOICES{$artist});		
	}
	
	rewrite($PREVIOUS_CHOICES_DATA_PATH, $new_data);
	print "done\n\n";

	print "Enjoy the music!\n";

	exit;

}



# 	my ($tag);
# 
# 	$tag = get_mp3tag($file, 1);
# 	my $id3v1_artist = $tag->{ARTIST};
# 	my $id3v1_title  = $tag->{TITLE};
# 
# 
# 
# 	$tag = get_mp3tag($file, 2);
# 	my $id3v2_artist = $tag->{ARTIST};
# 	my $id3v2_title  = $tag->{TITLE};



sub read_line_from_keyboard {
	my $key = <STDIN>;
	$key =~ s/[\r\n]//g;
	return $key;
}


sub pad_right {
	my($s, $t) = @_;
	return $s . (' ' x ($t - length($s)));
}

sub pad_left {
	my($s, $t) = @_;
	return (' ' x ($t - length($s))) . $s;
}


sub lc_sort
	{
	return sort {lc($a) cmp lc($b)} @_;
	}

sub split_path {
	my($full_path) = @_;
	$full_path =~ /^((.*)[\/\\])?([^\/\\]*)$/s;
	return ($2, $3);
}
	
	
sub directory_listing	{

	my ($dir_path, $attrib, $recursive, $dos_match, $cut_extension) = @_;
	      ## $atrib: 0: nur files, 1: nur directories, 2: files & directories
	
	$dos_match =~ s/([^0-9a-z\*\+\?])/\\$1/gsi;
	   
	my %regRef = ('*','.*','?','.','+','.+');
	$dos_match =~ s/([\*\?\+])/ $regRef{$1} /ge;

	my @rev;
	opendir(mydir,$dir_path);
	
	foreach my $entry (readdir(mydir)) {
		if ($entry !~ /^\.+$/) {
			my $matches = ($dos_match eq '' || $dos_match eq '.*\..*' || $entry =~ /^ $dos_match $/xi  );
			if (-f $dir_path . '/' . $entry && ($attrib == 0 || $attrib == 2) && $matches) {
				push(@rev, $entry);
			}
			if (-d $dir_path . '/' . $entry) {
				if ($attrib == 1 || $attrib == 2 && $matches) {
					push(@rev, $entry);
				}
				if ($recursive)	{
					push(@rev, map($_ = $entry . '/' . $_, directory_listing($dir_path . '/' . $entry, $attrib, $recursive)));
				}
			}
		}
	}

	closedir(mydir);
	
	if ($cut_extension) {
		map($_ =~ s/\.[^\.]*$//, @rev);
	}
	
	return @rev;	
}


sub read_file {
	my ($path) = @_;
	local *fh;
	open(fh, $path);
	my $rev = join ('', <fh>);
	close fh;
	return $rev;
}

sub rewrite {
	my($path, $text) = @_;
	local *fh;
	open (fh, ">$path");
	print fh $text;
	close fh;
}


sub error {
	my($msg) = @_;
	print "\n\nError: $msg\n";
	exit;
}


sub shrink
	{
	my ($string) = @_;
	$string =~ s/(^[\ \n\r\t]*|[\ \n\r\t]*$)//gsx;
	return $string;	
	}

sub lc_sort
	{
	return sort {lc($a) cmp lc($b)} @_;
	}
	
	
sub split_table	{
	my($table) = @_;
	return split(/\n/, $table);
}


sub escape_cell {
	my($cell) = @_;
	unless(defined $cell) { return "\\u"; }
	return parallel_replace($cell, 
		"\\",	"\\b",      
		"\n",	"\\n",
		"|" ,	"\\p",
		"\t",	"\\t",
		"\r",	"",
		"\u",	"",	# wie auch immer das da reingeraten koennte, nur zur sicherheit
		);
}
	

sub unescape_cell {
	my($cell) = @_;
	if($cell eq "\\u") { return undef }
	return parallel_replace($cell, 
		"\\b",	"\\",
		"\\n",	"\n",
		"\\p",	"|",
		"\\t",	"\t",
		"\t",	"",	# Tabs nur zur Prettiness
		);
}

sub split_column {
	my($table) = @_;
	# Konvertiert einen Table der genau 1 Zelle breit ist und X Zeilen hoch ist zu einem Array
	return map ( (split_row($_))[0], split(/\n/, $table));
}


sub make_column	{
	# Macht einen Table, der genau 1 Zelle breit ist und X Zeilen hoch ist
	return join('', map(escape_cell($_) . "\n", @_));
}


sub make_list {  # (1)(2)(23)(440)
	return join('', map('(' . parallel_replace($_, "(", "\\o", ")", "\\c") . ')', @_));
}
	
sub split_list {
	my($list) = @_;
	$list =~ s/^\(//;
	$list =~ s/\)$//;
	return map(parallel_replace($_, "\\o", "(", "\\c", ")"), split(/\)\(/, $list));
}


sub split_row {
	my ($value) = @_;
	strip_trailing_breaks($value);
	return map ( unescape_cell($_), split(/\|/, $value));
}
	

sub make_row {
	return join('|', map(escape_cell($_), @_)) . "\n";
}	

sub parallel_replace {
	my ($in,%ref) = (shift, @_);
	my $regex = join("|", map(quotemeta,keys %ref));
	$in =~ s/($regex)/$ref{$1}/gse;
	return $in;
}

sub strip_trailing_breaks {
	$_[0] =~ s/[\n\r]+$//;
}


